"""
Code that consumes PugSQL-dialect sql strings and returns validated
`pugsql.statement.Statement` objects.
"""

import re
import sys
from itertools import takewhile

from . import context, lexer, statement
from .exceptions import ParserError

_one = statement.One()
_many = statement.Many()
_affected = statement.Affected()
_scalar = statement.Scalar()
_insert = statement.Insert()
_raw = statement.Raw()


def parse(pugsql, ctx=None):
    """
    Processes the SQL string given in `pugsql` and returns a valid
    `pugsql.statement.Statement` object.

    Will raise a `pugsql.exceptions.ParserError` in any number of cases in
    which the PugSQL metadata isn't valid. However, this does not parse and
    validate the SQL statement.

    `ctx` is a context object provided by the `pugsql.context.Context`
    function, or `None`. If it is `None` a default context is created which
    will indicate that the SQL is being parsed from a literal string.
    """
    ctx = ctx or context.Context("<literal>")

    stream = lexer.lex(pugsql, ctx)
    leading_comments = _leading_comments(stream)
    rest = stream[len(leading_comments) :]

    cpr = _parse_comments(leading_comments)

    hdr = []
    if sys.version_info[:2] > (3, 9):
        hdr = ["-- pugsql function %s in file %s" % (cpr["name"], ctx.sqlfile)]
    sql = "\n".join(hdr + cpr["unconsumed"] + [token.value for token in rest])

    return statement.Statement(
        name=cpr["name"],
        sql=sql,
        doc=cpr["doc"],
        result=cpr["result"],
        filename=ctx.sqlfile if ctx.sqlfile != "<literal>" else None,
    )


def _leading_comments(stream):
    def is_comment(t):
        # allow blank whitespace lines in the leading comment
        return t.tag == "C" or t.value == "" or re.match(r"^\s+$", t.value)

    return list(takewhile(is_comment, stream))


def _parse_comments(comments):
    cpr = {
        "name": None,
        "result": _raw,
        "doc": None,
        "unconsumed": [],
    }

    for comment_token in comments:
        toks = lexer.lex_comment(comment_token)
        if not toks:
            cpr["unconsumed"].append(comment_token.value)
        elif toks["keyword"].value == ":name":
            _consume_name(cpr, toks["rest"])
        elif toks["keyword"].value == ":result":
            _consume_result(cpr, toks["rest"])
        else:
            cpr["unconsumed"].append(comment_token.value)

    return cpr


def _consume_result(cpr, rest):
    if not rest.value:
        raise ParserError("expected keyword", rest)
    _set_result(cpr, rest)


def _consume_name(cpr, rest):
    tokens = lexer.lex_name(rest)
    if not tokens:
        raise ParserError("expected a query name.", rest)

    name = tokens["name"].value
    if not _is_legal_name(name):
        raise ParserError(
            "'%s' is not a legal Python function name." % name, tokens["name"]
        )

    cpr["name"] = name

    if not tokens["keyword"].value:
        if tokens["rest"].value:
            raise ParserError(
                "encountered unexpected input after query name.",
                tokens["rest"],
            )
        return

    if tokens["rest"].value:
        raise ParserError(
            "encountered unexpected input after result type.", tokens["rest"]
        )

    _set_result(cpr, tokens["keyword"])


def _set_result(cpr, ktok):
    tokens = lexer.lex_result(ktok)
    if not tokens:
        raise ParserError("expected keyword", ktok)

    if tokens["rest"].value:
        raise ParserError(
            "encountered unexpected input after result type", tokens["rest"]
        )

    keyword = tokens["keyword"].value

    if keyword == ":one" or keyword == ":1":
        cpr["result"] = _one
    elif keyword == ":many" or keyword == ":*":
        cpr["result"] = _many
    elif keyword == ":affected" or keyword == ":n":
        cpr["result"] = _affected
    elif keyword == ":insert":
        cpr["result"] = _insert
    elif keyword == ":scalar":
        cpr["result"] = _scalar
    elif keyword != ":raw":
        raise ParserError("unrecognized keyword '%s'" % keyword, ktok)


def _is_legal_name(value):
    return re.match(r"^[a-zA-Z_][a-zA-Z0-9_]+$", value) is not None



"""
Functions that take strings and yield streams or dicts of `pugsql.lexer.Token`
objects, keeping track of source location.
"""

import re
from collections import namedtuple

from . import context

__pdoc__ = {}


Token = namedtuple("Token", ["tag", "value", "context"])
__pdoc__["Token"] = "A tagged string produced by the lexer."
__pdoc__["Token.tag"] = "A character indicating the meaning of the `value`."
__pdoc__["Token.value"] = "The string value of the `Token`."
__pdoc__["Token.context"] = (
    "A `pugsql.context.Context` for tracking source " "code information."
)


def lex(pugsql, ctx):
    """
    Splits the provided multiline PugSQL string into Tokens.
    """

    def generate(pugsql, ctx):
        for line in pugsql.splitlines():
            ctx = context.advance(ctx, lines=1)
            yield _categorize(line, ctx)

    return list(generate(pugsql, ctx))


def _categorize(line, ctx):
    line, ctx = _whitespace_advance(line, ctx)

    if line.startswith("--"):
        return Token("C", line, ctx)
    return Token("Q", line, ctx)


def lex_comment(token):
    m = re.match(
        r"(?P<lead>--+\s*)"
        r"(?P<keyword>\:[^ ]+)"
        r"(?P<internalws>\s+)?"
        r"(?P<rest>.*)?",
        token.value,
    )

    if not m:
        return None

    d = m.groupdict()
    restbegin = sum(len(d[k] or "") for k in d.keys() if k != "rest")

    return {
        "keyword": Token(
            "K",
            d["keyword"],
            context.advance(token.context, cols=len(d["lead"])),
        ),
        "rest": Token(
            "S", d["rest"], context.advance(token.context, cols=restbegin)
        ),
    }


def lex_name(token):
    line, ctx = _whitespace_advance(token.value, token.context)

    m = re.match(
        r"(?P<name>[^ ]+)"
        r"(?P<internalws>\s+)?"
        r"(?P<keyword>\:[^ ]+)?"
        r"(?P<internalws2>\s+)?"
        r"(?P<rest>.+)?",
        line,
    )

    if not m:
        return None

    d = m.groupdict()

    kwbegin = len(d["name"]) + len(d["internalws"] or "")
    kwctx = context.advance(ctx, cols=kwbegin)
    restbegin = sum(len(d[k] or "") for k in d.keys() if k != "rest")
    restctx = context.advance(ctx, cols=restbegin)
    return {
        "name": Token("N", d["name"], ctx),
        "keyword": Token("K", d["keyword"], kwctx),
        "rest": Token("S", d["rest"], restctx),
    }


def lex_result(token):
    line, ctx = _whitespace_advance(token.value, token.context)
    m = re.match(r"(?P<keyword>\:[^ ]+)" r"(?P<rest>.+)?", line)

    if not m:
        return None

    d = m.groupdict()
    restctx = context.advance(ctx, cols=len(d["keyword"]))

    return {
        "keyword": Token("K", d["keyword"], ctx),
        "rest": Token("S", d["rest"], restctx),
    }


def _whitespace_advance(line, ctx):
    ctx = context.advance(ctx, cols=len(line) - len(line.lstrip()))
    return line.strip(), ctx



"""
Code that processes SQL files and returns modules of database functions.
"""

import os
import re
import threading
from contextlib import contextmanager, suppress
from glob import glob

from sqlalchemy import create_engine
from sqlalchemy.exc import ResourceClosedError
from sqlalchemy.orm import sessionmaker

from . import context, parser
from .exceptions import NoConnectionError

__pdoc__ = {}


class Module(object):
    """
    Holds a set of SQL functions loaded from files.
    """

    sqlpaths = None
    engine = None

    def __init__(self, sqlpath, encoding=None):
        """
        Loads functions found in the *sql files specified by `sqlpath` into
        properties on this object. An `encoding` for the files can optionally
        be provided.

        The named sql functions in files should be unique.
        """
        self.sqlpaths = set()
        self._statements = {}
        self._sessionmaker = None
        self._locals = threading.local()

        self.add_queries(sqlpath, encoding=encoding)

    def add_queries(self, *paths, encoding=None):
        """
        Adds queries from *sql files in one or more `paths` to the module.
        An `encoding` for the files can optionally be provided.

        The named sql functions in files should be unique.
        """
        for p in paths:
            self._add_path(p, encoding=encoding)
        self.sqlpaths |= set(paths)

    def _add_path(self, sqlpath, encoding=None):
        if not os.path.isdir(sqlpath):
            raise ValueError("Directory not found: %s" % sqlpath)

        for sqlfile in sorted(glob(os.path.join(sqlpath, "*sql"))):
            with open(sqlfile, "r", encoding=encoding) as f:
                pugsql = f.read()

            # handle multiple statements per file
            statements = re.split(r"\n+(?=--+\s*:name)", pugsql)
            for statement in statements:
                s = parser.parse(statement, ctx=context.Context(sqlfile))

                if hasattr(self, s.name):
                    if s.name not in self._statements:
                        raise ValueError(
                            'Error loading %s - the function name "%s" is '
                            "reserved. Please choose another name."
                            % (sqlfile, s.name)
                        )
                    raise ValueError(
                        "Error loading %s - a SQL function named %s was "
                        "already defined in %s."
                        % (sqlfile, s.name, self._statements[s.name].filename)
                    )

                s.set_module(self)

                setattr(self, s.name, s)
                self._statements[s.name] = s

    @contextmanager
    def transaction(self):
        """
        Returns a session that manages a transaction scope, in which
        many statements can be run. Statements run on this module will
        automatically use this transaction. The normal use case  is to use this
        like a context manager, rather than interact with the result:

            foo = pugsql.module('sql/foo')
            with foo.transaction():
                x = foo.get_x(x_id=1234)
                foo.update_x(x_id=1234, x+1)

            # when the context manager exits, the transaction is committed.
            # if an exception occurs, it is rolled back.

        The transaction is active for statements executed on the current thread
        only.

        For engines that support SAVEPOINT, calling this method a second time
        begins a nested transaction.

        For more info, see here:
        https://docs.sqlalchemy.org/en/13/orm/session_transaction.html
        """
        if not getattr(self._locals, "session", None):
            if not self._sessionmaker:
                raise NoConnectionError()

            self._locals.session = self._sessionmaker()

            session = self._locals.session
            try:
                yield session
                session.commit()
            except Exception as e:
                session.rollback()
                raise e
            finally:
                session.close()
                self._locals.session = None
        else:
            session = self._locals.session.begin_nested()
            try:
                yield session
            except Exception as e:
                session.rollback()
                raise e
            else:
                with suppress(ResourceClosedError):
                    session.commit()

    def _execute(self, clause, *multiparams, **params):
        if getattr(self._locals, "session", None):
            if multiparams:
                return self._locals.session.execute(clause, *multiparams)
            else:
                return self._locals.session.execute(clause, params)

        if not self.engine:
            raise NoConnectionError()

        with self.engine.connect() as conn:
            if multiparams:
                result = conn.execute(clause, *multiparams)
            else:
                result = conn.execute(clause, params)
            conn.commit()
            return result

    @property
    def _dialect(self):
        """
        Gets the dialect for the SQLAlchemy engine.
        """
        if not self.engine:
            raise NoConnectionError()
        return self.engine.dialect

    def connect(self, connstr, **kwargs):
        """
        Sets the connection string for SQL functions on this module.

        See https://docs.sqlalchemy.org/en/13/core/engines.html for examples of
        legal connection strings for different databases.
        """
        self.setengine(create_engine(connstr, **kwargs))

    def setengine(self, engine):
        """
        Sets the SQLAlchemy engine for SQL functions on this module. This can
        be used instead of the connect method, when more customization of the
        connection engine is desired.

        See also: https://docs.sqlalchemy.org/en/13/core/connections.html
        """
        self.engine = engine
        self._sessionmaker = sessionmaker(bind=engine)

    def disconnect(self):
        """
        Disassociates the module from any connection it was previously given.
        """
        self.engine = None
        self._sessionmaker = None

    def __iter__(self):
        return iter(self._statements.values())


__pdoc__["Module.sqlpaths"] = (
    "A list of paths that the `pugsql.compiler.Module` was loaded from."
)
__pdoc__["Module.engine"] = (
    "The sqlalchemy engine object being used by the `pugsql.compiler.Module`."
)
